
import java.awt.Point;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Point2D;
import java.awt.geom.Ellipse2D.Double;
import java.awt.image.Raster;
import java.awt.image.WritableRaster;
import java.util.ArrayList;
import java.util.List;
import java.util.TreeMap;
import java.util.Vector;

import javax.swing.SwingWorker;


/**
 * Trova gli ellissi in un immagine con la Randomized Hough Transform.
 */

class RHTEllipse
	extends SwingWorker<List<Ellipse2D.Double>, Integer>
{	
	
    public static final int BLACK = 0;
    public static final int WHITE = 255;
    
    // parametri per il riconoscimento
    	// range del semi-asse maggiore
			private double  minA;
			private double  maxA;
			
		// range del semi-asse minore
			private double  minB;
			private double  maxB;
	
		// threshold - numero minimo di voti per il riconoscimento
			private int 	minVotes;
	
		// numero di iterazioni "idle" dopo le quali fermarsi
			private int 	idleStopSearch;
			
		// numero massimo di ellissi da riconoscere
			private int 	maxDetected;
	
		// immagine sorgente e di destinazione
            Raster src;
            WritableRaster dst;

	// costruttore che imposta tutti i parametri
	public RHTEllipse(Integer minMajorAxis,
			          Integer maxMajorAxis,
			          Integer minMinorAxis,
			          Integer maxMinorAxis,
			          Integer idleStop,
			          Integer maxDetected,
			          Raster src,
			          WritableRaster dst )
    {                
        this.minA = (double)minMajorAxis.intValue();
        this.minB = (double)minMinorAxis.intValue();
        this.maxA = (double)maxMajorAxis.intValue();
        this.maxB = (double)maxMinorAxis.intValue();
        //this.minVotes = 20; // TODO: rendere parametrico?
        this.minVotes = (int)(minA + minB);
        //this.minVotes = (int)(minB*2*Math.PI);
        this.idleStopSearch = idleStop.intValue();
        this.maxDetected = maxDetected.intValue();
        
        this.src = src;
        this.dst = dst;
    }

	// metodo che esegue il riconoscimento ed elabora un immagine contentente
	// gli edge points analizzati (in verde)
	@Override
	protected List<Ellipse2D.Double> doInBackground() throws Exception
    {
    	
    	// crea un vettore di punti contente tutti gli edge-points dell'immagine
	    	Vector<Point> _edge = new Vector<Point>();
	    	int width = src.getWidth();
	    	int heigth = src.getHeight();
	    	    	
	    	for(int i=0;i<width;i++)
	    	{
	    		for(int j=0;j<heigth;j++)
	    		{
	    			int sample = src.getSample(i,j,0);
	    			if(sample==1) // 1 is equal to white in binary images
	    				_edge.addElement(new Point(i,j));
	    		}
	    	}
    	
	    // se l'immagine non contiene edge points termina subito
	    	if(_edge.isEmpty())
	    	{
	    		System.out.println("No edge points detected");
	    		return null;
	    	}
    	
	    	
    	// ricava un array di edge-points dal vettore, così settato:
    	// edge[i][0] -> coordinata X del punto
    	// edge[i][1] -> coordinata Y del punto
    	// edge[i][2] -> 1=punto attivo, 0=punto scartato
	    	int edgePoints = _edge.size();   
	    	double edge[][];
	    	edge = new double[edgePoints][3];
	    	for(int i=0;i<edgePoints;i++)
	    	{
	    		edge[i][0] = ((Point)_edge.elementAt(i)).x;
	    		edge[i][1] = ((Point)_edge.elementAt(i)).y;
	    		edge[i][2] = 1;	// active point
	    	}
    	
    	if(maxB > maxA)
    		maxB = maxA;
    	
    	
    	// imposta il numero massimo di coppie = a ~50% del n° di edgepoints nell'immagine
    		int maxPairs = (int)((double)edgePoints * 0.5);
        	
        // crea l'accumulatore monodimensionale
	    // il parametro da votare è "b", il semi-asse minore
	        int accLength = (int)(this.maxA - this.minB); // la lunghezza dell'accumulatore, = alla cardinalità del dominio di "b"
	        int[] accumulator = new int[accLength];
	     	clear(accumulator); // inizializza l'accumulatore a 0
     	
	     	
        // variabili d'appoggio
	        int 	p1 		= 0; // primo punto della coppia di vertici
	        int 	p2 		= 0; // secondo punto della coppia di vertici
	        double 	dP1P2 	= 0.0; // distanza p1-p2
	        double 	cx		= 0.0; // coordinata x del centro
	        double 	cy		= 0.0; // coordinata y del centro
	        double 	a		= 0.0; // parametro dell'ellipse: semi-asse maggiore
			double 	d		= 0.0; // parametro dell'ellipse
			double 	f		= 0.0; // parametro dell'ellipse       
			double 	cos		= 0.0; // coseno dell'angolo dal centro al punto p3 (dichiarato in seguito)
			double 	b		= 0.0; // parametro dell'ellipse: semi-asse minore
			int idleStopCounter = 0; // contatore del numero di passi "idle"
			TreeMap<Integer, Ellipse2D> ellipsesVotes = new TreeMap<Integer, Ellipse2D>(); // mappa di ellissi ordinati secondo il n° di voti
			
		int currentPair = 0; // contatore del numero di coppie testate
        
		while(currentPair <= maxPairs) //termina se si è raggiunto il n° max di coppie
        {        
        	// ferma la ricerca se si è raggiunto il numero max di passi "idle"
        	if(idleStopCounter >= this.idleStopSearch)
        	{
        		System.out.println("idle stop");
        		break;
        	}
        	else idleStopCounter++;
        	       	
        	// seleziona casualmente una coppia di edge points attivi e distinti (p1, p2) 
	        	p1 = (int)Math.round(Math.random()*(double)(edgePoints-1));
	        	while(edge[p1][2]==0) // has been removed
	        		p1 = (int)Math.round(Math.random()*(double)(edgePoints-1));
	        	p2 = p1;        	
	        	while(p2==p1 || edge[p2][2]==0)
	        		p2 = (int)Math.round(Math.random()*(double)(edgePoints-1));
        	
        	// calcola la distanza tra di essi nell'immagine
	        	dP1P2 = Point2D.distance(edge[p1][0],edge[p1][1],edge[p2][0],edge[p2][1]);
        	
	        // controlla se la distanza è nel range settato per l'asse maggiore
	        if(!( dP1P2 >= 2*minA && dP1P2 <= 2*maxA ))
	        	continue; // altrimenti seleziona una nuova coppia
	            	
			currentPair++; // incrementa il numero di coppie testate
			this.setProgress(100*currentPair/maxPairs);
			
			// marca in verde i punti della coppia
			if(dst!=null) {
        		dst.setSample((int)edge[p1][0],(int)edge[p1][1],1,WHITE);
        		dst.setSample((int)edge[p2][0],(int)edge[p2][1],1,WHITE);
			}
    	
        	// calcola alcuni parametri dell'ellisse:
        		cx = (edge[p1][0] + edge[p2][0]) / 2.0; // centro X
        		cy = (edge[p1][1] + edge[p2][1]) / 2.0; // centro Y
        		a = dP1P2 / 2.0; // semi-asse maggiore
    		
        	// cicla tutti i restanti edge points ancora attivi
        	for(int p3 =0 ;p3<edgePoints ; p3++)
			{
				if(p3==p1 || p3==p2 || edge[p3][2]==0)
					continue;
				
				// calcola la distanza dal centro
				d = Point2D.distance(edge[p3][0],edge[p3][1],cx,cy);			
				
				// controlla se la distanza è nel range degli assi settato...
				if( d >= this.minB && d <= a )
				{
					// calcola i restanti parametri:
						f = Point2D.distance(edge[p3][0],edge[p3][1],edge[p2][0],edge[p2][1]);
						cos = (a*a + d*d - f*f) / (2*a*d);
						b = Math.sqrt( (a*a*d*d*(1-cos*cos)) / (a*a - d*d*cos*cos));
					
					if( b > maxB ) // scarta il punto se l'asse minore calcolato è troppo grande
						continue;
						
					// effettua la votazione nell'accumulatore
						int bIDX = (int)Math.round(b - minB);
						if(bIDX < 0 || bIDX >= accumulator.length) continue; // controllo range
						accumulator[bIDX]++; // effettua il voto
						
				} //endif		
			} // endfor
				
			
        	// ricerca 1 massimo nell'accumulatore
        	// ritorna un array: [0]->indice dove si è trovato il max, [1]->valore max
			int maxData[] = this.max(accumulator);
			int maxAccElement = maxData[1];
			int minorAxisLength = maxData[0] + (int)minB;
			   
			// controlla che il n° di voti sia compatibile con il threshold
			if( maxAccElement >= this.minVotes )
			{
				//costruisce l'ellisse con i parametri calcolati nella fase precedente
				// TODO? riconoscere l'orientamento dell'ellisse
				//if(houghMinorLength<a)
				Ellipse2D.Double ellipse = new Ellipse2D.Double( cx - minorAxisLength, // the X coordinate of the upper-left corner of the framing rectangle
                                                                 cy - a, // the Y coordinate of the upper-left corner of the framing rectangle
                                                                 minorAxisLength*2, // the width of the framing rectangle
                                                                 a*2 );  // the height of the framing rectangle

				ellipsesVotes.put(maxAccElement, ellipse);				
								
				idleStopCounter = 0; // azzera il contatore "idle"
				
			} //endif
					
			
			clear(accumulator); // ri-azzera l'accumulatore e seleziona una nuova coppia di vertici
	        
        } // endwhile
        

 // ULTIMA FASE: selezione del n° di ellissi richiesto
 
        this.setProgress(100);
     
		List<Ellipse2D.Double> ellipses = new ArrayList<Ellipse2D.Double>();
        
	     // prende il numero di ellissi richiesti con il max dei voti
		while(ellipses.size()!=maxDetected) {
			
			if(ellipsesVotes.isEmpty()) break;
			
			Ellipse2D.Double ellipse =
				(Ellipse2D.Double) ellipsesVotes.pollLastEntry().getValue();
			
			if(isValidEllipse(ellipse, ellipses))
				ellipses.add(ellipse);
			
			// TODO: marcare il centro ed i vertici in dst?
			//dst.setSample(_x,_y,0,WHITE);
			//System.out.println(ellipse.toString()); // stampa le caratteristiche dell'ellisse sullo schermo
		}
		
		System.out.println("requested: "+maxDetected+" ellipses");
        System.out.println("found: "+ellipses.size()+" ellipses");
        
		return(ellipses);
	}
    
    boolean isValidEllipse(Ellipse2D.Double e, List<Ellipse2D.Double> l) {
		
    	// controlla che l'ellisse rispetti le proporzioni tipiche della testa umana
    	double ellipseRatio = e.getHeight()/e.getWidth(); 
		if( ellipseRatio<1.25 || ellipseRatio>2 )
			return false;
		
		// se c'è un ellisse con numero maggiore di voti che lo contiene o lo interseca
		// l'ellisse corrente viene scartato (le teste sono necessariamente separate)
		if(l!=null)
			for(int j=0;j<l.size();j++) {
				if(l.get(j).getFrame().intersects(e.getFrame()))
				//if(l.get(j).contains(e.getFrame()))
					return false;
			}
			// ALTERNATIVA: confrontare le distanze dei centri
		
		return true;
    }
    
    
   	/**
   	 * ricerca un massimo in un array e ritorna un altro array:
     * [0]->indice dove si è trovato il max, [1]->valore max
   	 */
    private int[] max(int elements[])
    {
    	int max = 0;
    	int idx = -1;
    	for(int i=0;i<elements.length;i++)
    		if( elements[i] > max )
    		{
    			max = elements[i];
    			idx = i;
    		}
    	return new int[]{idx,max};
    }
    
    
    /**
     * azzera gli elementi di un array
     */
    private void clear(int elements[]) 
    {
    	java.util.Arrays.fill(elements, 0);
    }

}
